/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.lang.reflect.Array;

/**
 * An expression describing an array reference.
 *
 * @author Chris Schultz
 * @version $Revision: 1.2 $ $Date: 2008-06-20 22:58:48 $
 */
public class ArrayReferenceExpression
    extends Expression
{
    /**
     * The name of the array, or <code>&lt;anonymous&gt;</code> if the
     * array will be determined at evaluation-time.
     */
    private String _arrayName;

    /**
     * The expression that will evaluate to the array to be indexed.
     */
    private Expression _arrayExp;

    /**
     * The expression that identifies the index into the array.
     */
    private Expression _indexExp;

    /**
     * Creates a new ArrayReferenceExpression from the specified array
     * and index expressions.
     *
     * @param arrayExp The expression that will evaluate to the array to index.
     * @param indexExp The expression that will evaluate to the index of the
     *                 array.
     */
    public ArrayReferenceExpression(Expression arrayExp,
				    Expression indexExp)
    {
	if(null == arrayExp)
	    throw new IllegalArgumentException("Array must not be null.");
	if(null == indexExp)
	    throw new IllegalArgumentException("Index must not be null.");

	_arrayExp = arrayExp;
	_indexExp = indexExp;

	if(_arrayExp instanceof IdentifierExpression)
	    _arrayName = (String)((IdentifierExpression)_arrayExp)
		.getValue();
	else
	    _arrayName = "<anonymous>";
    }

    public void acceptVisitor(ExpressionVisitor visitor)
    {
	visitor.visitArrayReferenceExpression(this);
    }

    /**
     * Evaluates this Expression by evaluating the expression used for the
     * array index, and then returning the element of the array at that
     * location.
     *
     * @param ec The context in which this function call should be evaluated.
     *
     * @return The element in the array at the location specified by the index.
     *
     * @throws EvaluationException If there is an error evaluating the index
     *                             expression or locating the array element.
     */
    public Object evaluate(EvaluationContext ec)
	throws EvaluationException
    {
	Object array = _arrayExp.evaluate(ec);

	if(null == array)
	{
	    return null; // TODO: ?
	}
	// TODO: Allow Lists and other collections?
	else if(!array.getClass().isArray())
	{
	    throw new EvaluationException("Array reference '" + _arrayExp
					  + "' is not an array (it is"
					  + array.getClass() + "instead)");
	}

	Object indexValue = _indexExp.evaluate(ec);

	if(null == indexValue)
	    return null; // TODO: ?

	// TODO: accept LONG values within range?
	if(!(indexValue instanceof Integer))
	    throw new EvaluationException("Array index must be an int (it is "
					  + indexValue.getClass()
					  + " instead)");

	Object value = Array.get(array, ((Integer)indexValue).intValue());

	// Evaluate the object if it's an expression
	if(value instanceof Expression)
	    value = ((Expression)value).evaluate(ec);

	return value;
    }

    /**
     * Gets the expression that will evaluate to an array.
     */
    public Expression getArrayExpression()
    {
	return _arrayExp;
    }

    /**
     * Gets the expression that will evaluate to an index into the array.
     */
    public Expression getIndexExpression()
    {
	return _indexExp;
    }

    /**
     * Gets the name of the array, or <code>&lt;anonymous&gt;</code> if
     * the array will be determined at runtime.
     */
    public String getArrayName()
    {
	return _arrayName;
    }

    public List getSubExpressions()
    {
	ArrayList subs = new ArrayList(2);
	subs.add(_arrayExp);
	subs.add(_indexExp);

	return subs;
    }

    public String toString()
    {
	return _arrayName + "[ " + _indexExp + " ]";
    }
}
/*
set("o", "string")
set("a", call("getMethods", call("getClass", o)))
 */
