/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import java.util.Collections;
import java.util.List;

/**
 * An Expression which represents a fixed value (a constant).
 *
 * @author Chris Schultz
 * @version $Revision: 1.2 $ $Date: 2008-06-20 22:58:48 $
 */
public class ConstantExpression
    extends Expression
{
    /**
     * The value of this ConstantExpression.
     */
    private Object _value;

    /**
     * Creates a new ConstantExpression with the given value.
     *
     * @param value The value of the new ConstantExpressions.
     *
     * @throws IllegalArgumentException If the value is <code>null</code>.
     */
    public ConstantExpression(Object value)
    {
	if(null == value)
	    throw new IllegalArgumentException("Constant must not be null.");

	_value = value;
    }

    /**
     * Evaluates this Expression by returning its constant value.
     *
     * @param ec The EvaluationContext in which this expression should
     *           be evaluated.
     *
     * @return The value of this constant Expression.
     */
    public Object evaluate(EvaluationContext ec)
	throws EvaluationException
    {
	return _value;
    }

    /**
     * Returns a list containing a single element: this ConstantExpression.
     *
     * @return An immutable list containing this ConstantExpression.
     */
    public List getSubExpressions()
    {
	return Collections.singletonList(this);
    }

    /**
     * Gets the value of this constant Expression.
     *
     * @return The value of this constant Expression.
     */
    public Object getValue()
    {
	return _value;
    }

    /**
     * Returns the hash code for this ConstantExpression.
     *
     * The hash code of the ConstantExpression is equal to the
     * hash code of the value of this ConstantExpression.
     *
     * @return The hash code for this ConstantExpression.
     */
    public int hashCode()
    {
	return _value.hashCode();
    }

    /**
     * Returns a String representation of this Expression.
     *
     * @return A String representation of this Expression.
     */
    public String toString()
    {
	return "Constant " + (null == _value ? "<null>" : "<" + _value.getClass().getName() + ">" + String.valueOf(_value));
    }

    /**
     * Returns <code>true</code> if the specified object is equal to
     * this one.
     *
     * @param o The object to test for equality.
     *
     * @return <code>true</code> if the specified object is equal to
     *         this one, <code>false</code> otherwise.
     */
    public boolean equals(Object o)
    {
	return (this == o) ||
	    ((null != o)
	     && (this.getClass().equals(o.getClass()))
	     && (_value.equals(((ConstantExpression)o)._value)));
    }

    public void acceptVisitor(ExpressionVisitor visitor)
    {
	visitor.visitConstantExpression(this);
    }
}
