/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import java.util.Map;

/**
 * <p>
 *    Contains important information about the current expression
 *    evaluation.
 * </p>
 *
 * <p>
 *    The EvaluationContext stores all of the functions and constants that are
 *    bound and can be referenced by any expression. Everything is a
 *    first-class object in the evaluation context including functions,
 *    variables, etc. To bind an identifier to a value, see the {@link #set}
 *    method. To get the value of an identifier, see the {@link #get} method.
 * </p>
 *
 * @author Chris Schultz
 * @version $Revision: 1.10 $ $Date: 2008-06-20 22:58:48 $
 */
public interface EvaluationContext
{
    /**
     * Gets the value of the specified identifier.
     *
     * @param identifier The identifier for the value.
     *
     * @return The value of the specified identifier.
     */
    public Object get(String identifier);

    /**
     * Returns a Map of all bound identifiers and their values.
     * The Map is not modifiable.
     *
     * @return A Map of all identifiers as keys and their values
     *         as the values of the Map.
     */
    public Map getAll();

    /**
     * Sets the value of an identifer.
     *
     * @param identifier The identifier for the value.
     * @param value The value for the identifier.
     *
     * @return The <code>value</code> object.
     */
    public Object set(String identifier, Object value);

    /**
     * Removes the value of an identifier.
     *
     * @param identifier The identifier whose value should be removed.
     *
     * @return The old value of the object, or <code>null</code> if there was
     *         no value.
     */
    public Object remove(String identifier);

    /**
     * Sets the specified property to the given value.
     *
     * @param key The name of the property.
     * @param value The value of the property.
     */
    public void setProperty(String key, Object value);

    /**
     * Gets the specified property.
     *
     * @param key The name of the property.
     *
     * @return The value of the specified property.
     */
    public Object getProperty(String key);

    /**
     * Removes the specified property.
     *
     * @param key The name of the property to remove.
     */
    public void removeProperty(Object key);
}
