/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * An Expression representing a function call.
 *
 * @author Chris Schultz
 * @version $Revision: 1.7 $ $Date: 2008-06-20 22:58:48 $
 */
public class FunctionCallExpression
    extends Expression
{
    /**
     * An expression which returns the function to call.
     */
    private Expression _functionExp;

    /**
     * The name used to invoke this FunctionCallExpression, or
     * <code>&lt;anonymous&gt;</code> if the function will be determined
     * at evaluation-time.
     */
    private String _functionName;

    /**
     * A List of argument expressions.
     */
    private List _arguments;

    /**
     * <p>Creates a new FunctionCallExpression where the function to call
     * is determined at evaluation time.</p>
     */
    public FunctionCallExpression(Expression functionExp,
				  List arguments)
	throws EvaluationException
    {
 	if(null == functionExp)
 	    throw new IllegalArgumentException("Function expression must not be null.");

	_functionExp = functionExp;
	_arguments = arguments;

	_functionExp.setParent(this);

	if(null != arguments)
    {
	    for(Iterator i=arguments.iterator(); i.hasNext(); )
        {
            Expression exp = (Expression)i.next();

            if(null != exp)
                exp.setParent(this);
        }
    }

	if(_functionExp instanceof IdentifierExpression)
	    _functionName = (String)((IdentifierExpression)_functionExp)
		.getValue();
	else
	    _functionName = "<anonymous>";
    }

    /**
     * Returns the name of the function to be called by this
     * FunctionCallExpression, or <code>&lt;anonymous&gt;</code> if the
     * function will be determined at evaluation-time.
     *
     * @return The name of the function to be called by this
     *         FunctionCallExpression.
     */
    public String getFunctionName()
    {
	return _functionName;
    }

    /**
     * Gets the expression which will evaluate to the function to be called
     * by this FunctionCallExpression.
     *
     * @return The expression which, when evaluated, will return the Function
     * to be called by this FunctionCallExpression.
     */
    public Expression getFunctionExpression()
    {
	return _functionExp;
    }

    /**
     * Gets the list of argument expressions for this FunctionCallExpression.
     *
     * @return A List of Expressions objects which are the arguments
     *        to this FunctionCallExpression.
     */
    public List getArgumentExpressions()
    {
	return _arguments;
    }

    public List getSubExpressions()
    {
	ArrayList subs = new ArrayList(2 + (null == _arguments ? 0 : _arguments.size()));
	subs.add(this);
	subs.add(_functionExp);

	if(null != _arguments && 0 < _arguments.size())
	    for(Iterator i=_arguments.iterator(); i.hasNext(); )
		subs.addAll(((Expression)i.next()).getSubExpressions());

	return subs;
    }

    public void acceptVisitor(ExpressionVisitor visitor)
    {
	visitor.visitFunctionCallExpression(this);
    }

    /**
     * Evaluates this Expression by evaluating all of the arguments'
     * expressions, and then passing their resulting values to the
     * function bound to this FunctionCallExpression's name.
     *
     * @param ec The context in which this function call should be evaluated.
     *
     * @return The result of calling this FunctionVallExpression's function
     *         with it's arguments.
     *
     * @throws EvaluationException If there is an error evaluating any of
     *                             the argument expressions or calling the
     *                             function itself.
     */
    public Object evaluate(EvaluationContext ec)
	throws EvaluationException
    {
	Object func = _functionExp.evaluate(ec);

	if(null == func)
	    throw new EvaluationException("No function found for '" +
					  _functionExp + "'");
	else if(func instanceof Function)
	    return ((Function)func).call(ec,
					 _functionName,
					 _arguments);
	else
	    throw new EvaluationException("Invalid function: " + func);
    }

    /**
     * Returns a String representation of this FunctionCallExpression.
     *
     * @return A String representation of this FunctionCallExpression.
     */
    public String toString()
    {
	return _functionName
	    + "("
	    + (null == _arguments ? "" : String.valueOf(_arguments))
	    + ")"
	    ;
    }

    /**
     * Returns <code>true</code> if the specified object is equal to
     * this one.
     *
     * @param o The object to test for equality.
     *
     * @return <code>true</code> if the specified object is equal to
     *         this one, <code>false</code> otherwise.
     */
    public boolean equals(Object o)
    {
	if(this == o)
	    return true;

	if((null == o)
	   || (!this.getClass().equals(o.getClass())))
	    return false;

	FunctionCallExpression that = (FunctionCallExpression)o;

	return (this._functionExp.equals(that._functionExp))
	    && (this._arguments == that._arguments
		|| ((null == this._arguments)
		    && this._arguments.equals(that._arguments)));
    }

    public int hashCode()
    {
	int hashCode = super.hashCode(); // Just in case

	if(null != _functionExp)
	    hashCode = _functionExp.hashCode();
	else if(null != _functionName)
	    hashCode = _functionName.hashCode();

	if(null != _arguments)
	    for(Iterator i=_arguments.iterator(); i.hasNext(); )
		hashCode ^= i.next().hashCode();

	return hashCode;
    }
}
