/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import java.util.ArrayList;
import java.util.List;

/**
 * An Expression to perform an arithmetic sign inversion (unary minus) of a
 * sub-expression.
 *
 * @author Chris Schultz
 * @version $Revision: 1.4 $ $Date: 2008-06-20 22:58:48 $
 */
public class UnaryMinusExpression
    extends Expression
{
    /**
     * The expression whose result should be sign-inverted.
     */
    private Expression _sub;

    /**
     * Creates a new UnaryMinusExpression with the given sub-expression.
     *
     * @param subExpression The expression whose result will be sign inverted
     *                      when evaluated.
     *
     * @throws IllegalArgumentException If the sub expression is null.
     */
    public UnaryMinusExpression(Expression subExpression)
	throws EvaluationException
    {
	if(null == subExpression)
	    throw new IllegalArgumentException("Sub-expression of a negation must not be null.");

	// TODO: Put in an optimization to eliminate superfluous multiple
	// TODO: UnaryMinusExpressions ??

	_sub = subExpression;

	subExpression.setParent(this);
    }

    /**
     * Evaluates this Expression by returning the logical negation
     * of the sub-expression.
     *
     * @param ec The EvaluationContext in which this expression should
     *           be evaluated.
     *
     * @return The value of this UnaryMinusExpression.
     *
     * @throws EvaluationException If the sub-expression fails evaluation,
     *                             or does not return a Boolean value.
     */
    public Object evaluate(EvaluationContext ec)
	throws EvaluationException
    {
	Object subValue = _sub.evaluate(ec);

	if(null == subValue)
	    return null;
	else if(subValue instanceof Integer)
	    return new Integer(-((Integer)subValue).intValue());
	else if(subValue instanceof Double)
	    return new Double(-((Double)subValue).doubleValue());
	else if(subValue instanceof Long)
	    return new Long(-((Long)subValue).longValue());
	else if(subValue instanceof Float)
	    return new Float(-((Float)subValue).floatValue());
	else if(subValue instanceof Short)
	    return new Short((short)-((Short)subValue).shortValue());
	else if(subValue instanceof Byte)
	    return new Byte((byte)-((Byte)subValue).byteValue());
	else
	    throw new EvaluationException("Illegal numeric value: " + subValue);
    }

    /**
     * Gets the Expression which is being sign inverted by this one.
     *
     * @return The Expression which is being sign inverted by this one.
     */
    public Expression getSubExpression()
    {
	return _sub;
    }

    /**
     * Returns a list containing this UnaryMinusExpression and the expression
     * to be sign inverted (in that order).
     *
     * @return A list containing this UnaryMinusExpression and the expression
     *         to be sign inverted (in that order).
     */
    public List getSubExpressions()
    {
	ArrayList subs = new ArrayList(2);
	subs.add(this);
	subs.addAll(_sub.getSubExpressions());

	return subs;
    }

    public void acceptVisitor(ExpressionVisitor visitor)
    {
	visitor.visitUnaryMinusExpression(this);
    }

    /**
     * Returns the hash code for this UnaryMinusExpression.
     *
     * The hash code of the UnaryMinusExpression is equal to the
     * hash code of the value of this UnaryMinusExpression.
     *
     * @return The hash code for this UnaryMinusExpression.
     */
    public int hashCode()
    {
	return -(_sub.hashCode());
    }

    /**
     * Returns a String representation of this UnaryMinusExpression.
     *
     * @return A String representation of this UnaryMinusExpression.
     */
    public String toString()
    {
	return "-(" + _sub + ")";
    }

    /**
     * Returns <code>true</code> if the specified object is equal to
     * this one.
     *
     * @param o The object to test for equality.
     *
     * @return <code>true</code> if the specified object is equal to
     *         this one, <code>false</code> otherwise.
     */
    public boolean equals(Object o)
    {
	return (this == o)
	    || ((null != o)
		&& (this.getClass().equals(o.getClass()))
		&& (_sub.equals(((UnaryMinusExpression)o)._sub)));
    }
}
