/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator.function;

import java.util.List;

import net.christopherschultz.evaluator.EvaluationContext;
import net.christopherschultz.evaluator.EvaluationException;
import net.christopherschultz.evaluator.Function;
import net.christopherschultz.evaluator.StandardFunction;
import net.christopherschultz.evaluator.UnexpectedArgumentException;

/**
 * <p>
 *    Implements an if-then-else construct as a function.
 * </p>
 *
 * <p>
 *    Usage: if(predicate, then-expression, else-expression)
 * </p>
 *
 * <p>
 *    This function returns the result of the <i>then-expression</i>
 *    if the predicate evaluates to <code>true</code>, or returns
 *    the result of the <i>else-expression</i> if <i>predicate</i>
 *    evaluates to <code>false</code>. <code>null</code> is returned
 *    if the predicate returns <code>null</code>.
 * </p>
 *
 * @author Chris Schultz
 * @version $Revision: 1.2 $ $Date: 2008-06-20 22:58:48 $
 */
public class IfFunction
    implements Function
{
    /**
     * Creates a new IfFunction object.
     */
    public IfFunction()
    {
        super();
    }

    public Object call(EvaluationContext ec,
		       String functionName,
		       List argumentExpressions)
	throws EvaluationException
    {
	if(null == argumentExpressions || 3 != argumentExpressions.size())
	    throw new EvaluationException("Incorrect number of arguments. Expected 3, got " + (null == argumentExpressions ? "0" : String.valueOf(argumentExpressions.size())));

	Object predicate = StandardFunction
	    .getExpressionValue(ec, argumentExpressions.get(0));

	// Bail-out of the result is null.
	if(null == predicate)
	    return null;
	
	if(!(predicate instanceof Boolean))
	    throw new UnexpectedArgumentException("Expected boolean predicate, not "
						  + predicate.getClass());

	if(((Boolean)predicate).booleanValue())
	{
	    return StandardFunction
		.getExpressionValue(ec, argumentExpressions.get(1));
	}
	else
	{
	    return StandardFunction
		.getExpressionValue(ec, argumentExpressions.get(2));
	}
    }

    public String toString()
    {
	return "IfFunction";
    }
}
