/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator.util;

import java.util.Enumeration;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.ResourceBundle;

import net.christopherschultz.evaluator.EvaluationContext;
import net.christopherschultz.evaluator.Function;

/**
 * Contains utility methods for use with EvaluationContext objects.
 *
 * @author Chris Schultz
 * @version $Revision: 1.5 $ $Date: 2008-06-20 22:58:48 $
 */
public class ContextUtil
{
    private ContextUtil() { }

    /**
     * Loads constants, properties, and functions from the specified
     * ResourceBundle into the specified EvaluationContext.
     *
     * @param bundle The bundle containing all of the constant, property,
     *               and function definitions.
     * @param ec The EvaluationContext into which all the constants,
     *           properties, and functions will be placed.
     */
    public static void load(ResourceBundle bundle, EvaluationContext ec)
    {
        load(toMap(bundle), ec);
    }
    
    /**
     * Loads constants, properties, and functions from the specified
     * Map into the specified EvaluationContext.
     *
     * @param props The property map containing all of the constant, property,
     *               and function definitions.
     * @param ec The EvaluationContext into which all the constants,
     *           properties, and functions will be placed.
     */
    public static void load(Map props, EvaluationContext ec)
    {
	ConstantLoader loader = new ConstantLoader();

	for(Iterator i=props.entrySet().iterator(); i.hasNext(); )
	{
	    Map.Entry entry = (Map.Entry)i.next();
	    String key = (String)entry.getKey();
	    String keyValue = (String)entry.getValue();

	    if(key.startsWith("constant."))
	    {
		String constantName = key.substring(9);

		Object value = loader.getValue(keyValue);

		ec.set(constantName, value);
	    }
	    else if(key.startsWith("function."))
	    {
		String functionName = key.substring(9);

		ec.set(functionName, loader.getObject(keyValue));
	    }
	    else if(key.startsWith("property."))
	    {
		String propName = key.substring(9);

		// Is property name is really something else?
		//
		// First, resolve the property name into a string
		// (by looking up it's value using reflection)
		// and then use that string as the property's name.
		if(0 <= propName.indexOf('#'))
		    propName = (String)loader.getValue(propName);

		Object value = loader.getValue(keyValue);

		ec.setProperty(propName, value);
	    }
	}
    }

    /**
     * Converts a ResourceBundle into a Map.
     *
     * @param props A ResourceBundle.
     *
     * @return A Map containing all keys and values from the ResourceBoudle.
     */
    public static Map toMap(ResourceBundle props)
    {
	HashMap map = new HashMap();
	for(Enumeration e = props.getKeys(); e.hasMoreElements(); )
	{
	    String key = (String)e.nextElement();
	    map.put(key, props.getObject(key));
	}

	return map;
    }
}
