/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import java.util.ArrayList;
import java.util.Collections;

import net.christopherschultz.evaluator.IdentifierExpression;
import net.christopherschultz.evaluator.function.BinaryOperator;
import net.christopherschultz.evaluator.util.DefaultEvaluationContext;

/**
 * UnitTest for array reference expressions.
 *
 * @author Chris Schultz
 * @version $Revision: 1.5 $ $Date: 2008-06-20 22:58:48 $
 */
public class ArrayReferenceExpressionTests
    extends TestCase
{
    EvaluationContext ec = new DefaultEvaluationContext();

    public void setUp()
    {
        ec.set("+", new BinaryOperator.Add());
        ec.set("get", new net.christopherschultz.evaluator.function.GetVariableFunction());
    }

    public void testSimpleArray()
        throws Exception
    {
        ArrayReferenceExpression are = new ArrayReferenceExpression(new IdentifierExpression("a"),
                                                                    new ConstantExpression(new Integer(2)));

        Integer[] a = new Integer[3];
        for(int i=0; i<3; ++i)
            a[i] = new Integer(i);

        ec.set("a", a);

        Assert.assertEquals("Array reference doesn't work :(",
                            new Integer(2),
                            are.evaluate(ec));
    }

    public void testPrimitiveArray()
        throws Exception
    {
        ArrayReferenceExpression are = new ArrayReferenceExpression(new IdentifierExpression("a"),
                                                                    new ConstantExpression(new Integer(2)));

        int[] a = new int[3];
        for(int i=0; i<3; ++i)
            a[i] = i;

        ec.set("a", a);

        Assert.assertEquals("Primitive boxing fails.",
                            new Integer(2),
                            are.evaluate(ec));
    }

    public void testOutOfBounds()
        throws Exception
    {
        ArrayReferenceExpression are = new ArrayReferenceExpression(new IdentifierExpression("a"),
                                                                    new ConstantExpression(new Integer(2)));

        int[] a = new int[0];

        ec.set("a", a);

        try
        {
            are.evaluate(ec);
            Assert.fail("Expected ArrayIndexOutOfBoundsException");
        }
        catch (ArrayIndexOutOfBoundsException aioobe)
        {
            // Expected result
        }
    }

    public void testDynamicIndex()
        throws Exception
    {
        ArrayReferenceExpression are = new ArrayReferenceExpression(new IdentifierExpression("a"),
                                                                    new BinaryOperatorExpression(new ConstantExpression(new Integer(1)), "+", new ConstantExpression(new Integer(1))));

        int[] a = new int[3];
        for(int i=0; i<3; ++i)
            a[i] = i;

        ec.set("a", a);

        Assert.assertEquals("Primitive boxing fails.",
                            new Integer(2),
                            are.evaluate(ec));
    }

    public void testDynamicArrayAndIndex()
        throws Exception
    {
        // get("a")[1+1]
        ArrayReferenceExpression are = new ArrayReferenceExpression(new FunctionCallExpression(new IdentifierExpression("get"), Collections.singletonList(new IdentifierExpression("a"))),
                                                                    new BinaryOperatorExpression(new ConstantExpression(new Integer(1)), "+", new ConstantExpression(new Integer(1))));

        int[] a = new int[3];
        for(int i=0; i<3; ++i)
            a[i] = i;

        ec.set("a", a);

        Assert.assertEquals("Dynamic array evalution fails.",
                            new Integer(2),
                            are.evaluate(ec));
    }

    /**
     * Creates a new UnitTest to execute the specified test.<p>
     *
     * @param name The name of the test method to run.
     */
    public ArrayReferenceExpressionTests(String name)
    {
        super(name);
    }
}
