/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator.function;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import net.christopherschultz.evaluator.EvaluationContext;
import net.christopherschultz.evaluator.EvaluationException;
import net.christopherschultz.evaluator.Expression;
import net.christopherschultz.evaluator.parser.ExpressionParser;
import net.christopherschultz.evaluator.parser.ParseException;
import net.christopherschultz.evaluator.util.DefaultEvaluationContext;

/**
 * Unit test for the Heaviside (unit step) function.<p>
 *
 * @author Chris Schultz
 * @version $Revision: 1.3 $ $Date: 2008-06-20 22:58:48 $
 */
public class HeavisideTests
    extends TestCase
{
    EvaluationContext ec;

    private static final Integer ZERO = new Integer(0);
    private static final Integer ONE = new Integer(1);

    public void setUp()
    {
	ec = new DefaultEvaluationContext();

	ec.set("H", new MathFunction.heaviside());
	ec.set("=", new BinaryOperator.Equal());
	ec.set("create", new NewObjectFunction());
	ec.set("true", Boolean.TRUE);
	ec.set("false", Boolean.FALSE);
    }

    private Object evaluate(String expression,
			    EvaluationContext ec)
	throws EvaluationException, ParseException
    {
	Expression exp = ExpressionParser.parseExpression(expression);

	return exp.evaluate(ec);
    }

    public void testSimpleFalse()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(false)", ec));
    }

    public void testSimpleTrue()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(true)", ec));
    }

    public void testComplexFalse()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(0 = 1)", ec));
    }

    public void testComplexTrue()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(0 = 0)", ec));
    }

    public void testNull()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(null)", ec));
    }

    //
    // Test short values
    //
    public void testShortNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Short\", \"-1\"))", ec));
    }

    public void testShortZero()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Short\", \"0\"))", ec));
    }

    public void testShortPositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(create(\"java.lang.Short\", \"1\"))", ec));
    }

    //
    // Test integer values
    //
    public void testIntegerNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(-1)", ec));
    }

    public void testIntegerZero()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(0)", ec));
    }

    public void testIntegerPositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(1)", ec));
    }

    //
    // Test long values
    //
    public void testLongNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Long\", \"-1\"))", ec));
    }

    public void testLongZero()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Long\", \"0\"))", ec));
    }

    public void testLongPositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(create(\"java.lang.Long\", \"1\"))", ec));
    }

    //
    // Test float values
    //
    public void testFloatNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Float\", \"-1.0\"))", ec));
    }

    public void testFloatZero()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(create(\"java.lang.Float\", \"0.0\"))", ec));
    }

    public void testFloatPositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(create(\"java.lang.Float\", \"1.0\"))", ec));
    }

    //
    // Test double values
    //
    public void testDoubleNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(-1.0)", ec));
    }

    public void testDoubleZero()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(0.0)", ec));
    }

    public void testDoublePositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(1.0)", ec));
    }

    public void testDoubleTinyNegative()
	throws Exception
    {
	Assert.assertEquals(ZERO,
			    evaluate("H(-0.000000000000000000000000000000001)",
				     ec));
    }

    public void testDoubleTinyPositive()
	throws Exception
    {
	Assert.assertEquals(ONE,
			    evaluate("H(0.000000000000000000000000000000001)",
				     ec));
    }

    /**
     * Creates a new UnitTest to execute the specified test.<p>
     *
     * @param name The name of the test method to run.
     */
    public HeavisideTests(String name)
    {
        super(name);
    }
}
