/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator.function;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestSuite;

import org.jmock.MockObjectTestCase;
import org.jmock.Mock;

import java.util.ArrayList;

import net.christopherschultz.evaluator.*;
import net.christopherschultz.evaluator.util.DefaultEvaluationContext;

/**
 * UnitTest for expressions and evaluating them.<p>
 *
 * @author Chris Schultz
 * @version $Revision: 1.3 $ $Date: 2008-06-20 22:58:48 $
 */
public class NonShortCircuitAndTests
    extends net.christopherschultz.evaluator.UnitTest
{
    EvaluationContext ec;
    Expression exp;
    Mock f_true;
    Mock f_false;
    Object result;

    public void setUp()
    {
        f_true = new Mock(Function.class);
        f_false = new Mock(Function.class);

        ec = new DefaultEvaluationContext();

        ec.set("true", (Function)f_true.proxy());
        ec.set("false", (Function)f_false.proxy());
        ec.set("&&", new NonShortCircuitAnd());
    }

    public void testTestTest() { }

    public void testEvaluation()
        throws EvaluationException
    {
        f_true.stubs().will(returnValue(Boolean.TRUE));
        f_false.stubs().will(returnValue(Boolean.FALSE));

        exp = new BinaryOperatorExpression(new FunctionCallExpression(id("true"), null),
                                           "&&",
                                           new FunctionCallExpression(id("true"), null));

        result = exp.evaluate(ec);

        Assert.assertNotNull("Short-circuit AND results in null", result);
        Assert.assertEquals("Short-circuit logical does not return a boolean",
                            Boolean.class, result.getClass());
        Assert.assertEquals("AND doesn't evaluate properly",
                            Boolean.TRUE, result);

        exp = new BinaryOperatorExpression(new FunctionCallExpression(id("true"), null),
                                           "&&",
                                           new FunctionCallExpression(id("false"), null));

        result = exp.evaluate(ec);

        Assert.assertNotNull("Short-circuit AND results in null", result);
        Assert.assertEquals("Short-circuit logical does not return a boolean",
                            Boolean.class, result.getClass());
        Assert.assertEquals("AND doesn't evaluate properly",
                            Boolean.FALSE, result);

        exp = new BinaryOperatorExpression(new FunctionCallExpression(id("false"), null),
                                           "&&",
                                           new FunctionCallExpression(id("true"), null));

        result = exp.evaluate(ec);

        Assert.assertNotNull("Short-circuit AND results in null", result);
        Assert.assertEquals("Short-circuit logical does not return a boolean",
                            Boolean.class, result.getClass());
        Assert.assertEquals("AND doesn't evaluate properly",
                            Boolean.FALSE, result);

        exp = new BinaryOperatorExpression(new FunctionCallExpression(id("false"), null),
                                           "&&",
                                           new FunctionCallExpression(id("false"), null));

        result = exp.evaluate(ec);

        Assert.assertNotNull("Short-circuit AND results in null", result);
        Assert.assertEquals("Short-circuit logical does not return a boolean",
                            Boolean.class, result.getClass());
        Assert.assertEquals("AND doesn't evaluate properly",
                            Boolean.FALSE, result);
    }

    public void testNoShortCircuit()
        throws EvaluationException
    {
        f_true.expects(once()).method("call").with(ANYTHING, ANYTHING, ANYTHING).will(returnValue(Boolean.TRUE));
        f_false.expects(once()).method("call").with(ANYTHING, ANYTHING, ANYTHING).will(returnValue(Boolean.FALSE));

        exp = new BinaryOperatorExpression(new FunctionCallExpression(id("false"), null),
                                           "&&",
                                           new FunctionCallExpression(id("true"), null));

        result = exp.evaluate(ec);

        Assert.assertNotNull("Short-circuit AND results in null", result);
        Assert.assertEquals("Short-circuit logical does not return a boolean",
                            Boolean.class, result.getClass());
        Assert.assertEquals("AND doesn't evaluate properly",
                            Boolean.FALSE, result);
    }

    public void testShortCircuitNulls()
        throws EvaluationException
    {
        f_true.expects(never()).method("call").with(ANYTHING, ANYTHING, ANYTHING).will(returnValue(Boolean.TRUE));

        exp = new BinaryOperatorExpression(new IdentifierExpression("null"),
                                           "&&",
                                           new FunctionCallExpression(id("true"), null));

        result = exp.evaluate(ec);

        Assert.assertNull("Short-circuit AND does not result in null", result);
    }

    public void testNulls()
        throws EvaluationException
    {
        exp = new BinaryOperatorExpression(new IdentifierExpression("null"),
                                           "&&",
                                           new IdentifierExpression("null"));

        Assert.assertNull("Anding nulls does not return null",
                          exp.evaluate(ec));
    }

    public void testLeftNullTrue()
        throws EvaluationException
    {
        exp = new BinaryOperatorExpression(new IdentifierExpression("null"),
                                           "&&",
                                           new ConstantExpression(Boolean.TRUE));
        
        Assert.assertNull("Anding left null does not return null",
                          exp.evaluate(ec));
    }

    public void testLeftNullFalse()
        throws EvaluationException
    {
        exp = new BinaryOperatorExpression(new IdentifierExpression("null"),
                                           "&&",
                                           new ConstantExpression(Boolean.FALSE));
        
        Assert.assertNull("Anding left null does not return null",
                          exp.evaluate(ec));
    }

    public void testRightNullTrue()
        throws EvaluationException
    {
        exp = new BinaryOperatorExpression(new ConstantExpression(Boolean.TRUE),
                                           "&&",
                                           new IdentifierExpression("null"));
        
        Assert.assertNull("Anding right null does not return null",
                          exp.evaluate(ec));
    }

    public void testRightNullFalse()
        throws EvaluationException
    {
        exp = new BinaryOperatorExpression(new ConstantExpression(Boolean.FALSE),
                                           "&&",
                                           new IdentifierExpression("null"));
        
        Assert.assertNull("Anding right null does not return null",
                          exp.evaluate(ec));
    }

    public void testMultipleArgumentsMixedNulls()
        throws Exception
    {
        // Basically, any FALSE should get us FALSE.
        // Also, if there are nulls and no FALSEs,
        // we expect null.

        Function and = (Function)ec.get("&&");

        ArrayList args = new ArrayList();
        args.add(null);
        args.add(Boolean.FALSE); // Should get us FALSE
        args.add(Boolean.TRUE);

        Assert.assertNull("Null arguments should yield null result",
                          and.call(ec, "&&", args));

        args.clear();
        args.add(null);
        args.add(Boolean.TRUE);

        Assert.assertNull("Null arguments should yield null result",
                          and.call(ec, "&&", args));

        args.clear();
        args.add(Boolean.TRUE);

        Assert.assertEquals("Sufficient arguments to AND returns no value",
                            Boolean.TRUE, and.call(ec, "&&", args));
    }

    /**
     * Returns a Test which will execute all the test methods found
     * in this class.<p>
     *
     * @return A Test for JUnit to run.
     */
    public static Test suite()
    {
        TestSuite ts = new TestSuite(NonShortCircuitAndTests.class);

        return ts;
    }

    /**
     * Creates a new UnitTest to execute the specified test.<p>
     *
     * @param name The name of the test method to run.
     */
    public NonShortCircuitAndTests(String name)
    {
        super(name);
    }
}
