/**
 * This software is Copyright (C) 2008 Christopher Schultz
 *
 * Please see the COPYING file for more information.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 * 
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA
 */
package net.christopherschultz.evaluator.parser;

import junit.framework.Assert;
import junit.framework.Test;
import junit.framework.TestCase;
import junit.framework.TestSuite;

import net.christopherschultz.evaluator.*;
import net.christopherschultz.evaluator.parser.ExpressionParser;

/**
 * UnitTest for the ANTLR-generated parser.<p>
 *
 * @author Chris Schultz
 * @version $Revision: 1.4 $ $Date: 2009-04-09 16:02:58 $
 */
public class UnitTest
    extends TestCase
{
    private static ConstantExpression constant(int i)
    {
        return new ConstantExpression(new Integer(i));
    }

    private static ConstantExpression constant(boolean b)
    {
        return new ConstantExpression(b ? Boolean.TRUE : Boolean.FALSE);
    }

    private static ConstantExpression constant(Object o)
    {
        return new ConstantExpression(o);
    }

    private ExpressionParser parser = new ExpressionParser();

    public void testNumericLiterals()
        throws ParseException, EvaluationException
    {
        Assert.assertEquals("Parser fails for int atoms",
                            constant(1),
                            parser.parse("1"));

        Assert.assertEquals("Parser fails for signed int atoms",
                            constant(1),
                            parser.parse("+1"));

        Assert.assertEquals("Parser fails for signed int atoms",
                            new UnaryMinusExpression(new ConstantExpression(new Integer(1))),
                            parser.parse("-1"));

        Assert.assertEquals("Parser fails for double atoms",
                            constant(new Double(1.0)),
                            parser.parse("1.0"));

        Assert.assertEquals("Parser fails for signed double atoms",
                            constant(new Double(1.0)),
                            parser.parse("+1.0"));

        Assert.assertEquals("Parser fails for signed double atoms",
                            new UnaryMinusExpression(new ConstantExpression(new Double(1.0))),
                            parser.parse("-1.0"));

        Assert.assertEquals("Parser fails for decimal-led double atoms",
                            constant(new Double(0.5)),
                            parser.parse(".5"));

        Assert.assertEquals("Parser fails for signed decimal-led double atoms",
                            constant(new Double(0.5)),
                            parser.parse("+.5"));

        Assert.assertEquals("Parser fails for signed decimal-led double atoms",
                            new UnaryMinusExpression(new ConstantExpression(new Double(0.5))),
                            parser.parse("-.5"));
    }

    public void testHexLiterals()
        throws Exception
    {
        Assert.assertEquals("Parser fails for byte hex literals",
                            new ConstantExpression(new Byte((byte)0xf)),
                            parser.parse("0xf"));

        Assert.assertEquals("Parser fails for short hex literals",
                            new ConstantExpression(new Short((short)0x1ef)),
                            parser.parse("0x1ef"));

        Assert.assertEquals("Parser fails for int hex literals",
                            new ConstantExpression(new Integer(0x1a2b3c4d)),
                            parser.parse("0x1a2b3c4d"));

        Assert.assertEquals("Parser fails for long hex literals",
                            new ConstantExpression(new Long(0x1a2b3c4d5e6fl)),
                            parser.parse("0x1a2b3c4d5e6f"));
    }

    public void testOctalLiterals()
        throws Exception
    {
        Assert.assertEquals("Parser fails for octal byte literals",
                            new ConstantExpression(new Byte((byte)077)),
                            parser.parse("077"));

        Assert.assertEquals("Parser fails for octal short literals",
                            new ConstantExpression(new Short((short)0777)),
                            parser.parse("0777"));

        Assert.assertEquals("Parser fails for octal int literals",
                            new ConstantExpression(new Integer(0777777)),
                            parser.parse("0777777"));

        Assert.assertEquals("Parser fails for octal long literals",
                            new ConstantExpression(new Long(07777777777777l)),
                            parser.parse("07777777777777"));
    }

    public void testStringLiterals()
        throws Exception
    {
        Assert.assertEquals("Parser fails for string atoms",
                            constant("test"),
                            parser.parse("\"test\""));

        Assert.assertEquals("Parser fails for some strings",
                            constant("a,b,c"),
                            parser.parse("\"a,b,c\""));

        Assert.assertEquals("Parser fails for punctuation in strings",
                            constant(",./<>?[]{}|`~!@#$%^&*()_+-="),
                            parser.parse("\",./<>?[]{}|`~!@#$%^&*()_+-=\""));

        Assert.assertEquals("Parser fails for escaped strings",
                            constant("\"\n\r\t\b\f\\'"),
                            parser.parse("\"\\\"\\n\\r\\t\\b\\f\\\\\\\'\""));

        Assert.assertEquals("Parser fails for identifiers",
                            new IdentifierExpression("true"),
                            parser.parse("true"));
    }

    public void testSimpleExpressions()
        throws ParseException, EvaluationException
    {
        Assert.assertEquals("Parser fails for simple expressions",
                            new BinaryOperatorExpression(constant(2),
                                                         "+",
                                                         constant(2)),
                            parser.parse("2+2"));

        Assert.assertEquals("Parser fails for simple expressions",
                            new NegatedExpression(new BinaryOperatorExpression(constant(2),
                                                                               "=",
                                                                               constant(2))),
                            parser.parse("!(2 = 2)"));
    }

    public void testSubExpressions()
        throws ParseException, EvaluationException
    {
        Assert.assertEquals("Parser fails with sub-expressions",
                            constant(4),
                            parser.parse("(4)"));

        Assert.assertEquals("Parser fails with multiple sub-expressions",
                            constant(4),
                            parser.parse("((((4))))"));

        Assert.assertEquals("Parser fails with negated sub-expressions",
                            new NegatedExpression(constant(4)),
                            parser.parse("!(4)"));
    }

    public void testOrderOfOperations()
        throws ParseException, EvaluationException
    {
        Assert.assertEquals("Parser fails to correctly order addition and multiplication",
                            new BinaryOperatorExpression(constant(1),
                                                         "+",
                                                         new BinaryOperatorExpression(constant(2),
                                                                                      "*",
                                                                                      constant(3))),
                            parser.parse("1+2*3"));

        Assert.assertEquals("Parser fails to correctly order multiplication and addition",
                            new BinaryOperatorExpression(new BinaryOperatorExpression(constant(1),
                                                                                      "*",
                                                                                      constant(2)),
                                                         "+",
                                                         constant(3)),
                            parser.parse("1*2+3"));

        Assert.assertEquals("Parser fails order of operations for negation and relations",
                            new BinaryOperatorExpression(new NegatedExpression(constant(2)),
                                                         "=",
                                                         constant(2)),
                            parser.parse("! 2 = 2"));

        Assert.assertEquals("Parser fails to correctly order logical and relative expressions",
                            new BinaryOperatorExpression(new BinaryOperatorExpression(constant(2),
                                                                                      "=",
                                                                                      constant(2)),
                                                         "||",
                                                         new BinaryOperatorExpression(constant(3),
                                                                                      "=",
                                                                                      constant(3))),
                            parser.parse("2 = 2 || 3 = 3"));

        //
        // Just for the heck of it...
        //

        Assert.assertEquals("Parser failed order-of-operations",
                            parser.parse("1+(2*3)"),
                            parser.parse("1+2*3"));

        Assert.assertEquals("Parser failed order-of-operations",
                            parser.parse("(2 = 2) && (3 = 3)"),
                            parser.parse("2 = 2 && 3 = 3"));

        Assert.assertEquals("Parser failed order-of-operations",
                            parser.parse("(!2) && 3"),
                            parser.parse("! 2 && 3"));

        Assert.assertEquals("Parser failed order-of-operations",
                            parser.parse("(!2) && (!3)"),
                            parser.parse("! 2 && ! 3"));
    }

    public void testUnaryExpressions()
        throws ParseException, EvaluationException
    {
        Assert.assertEquals("Parser fails for embedded unary expressions",
                            new BinaryOperatorExpression(constant(2),
                                                         "+",
                                                         new UnaryMinusExpression(new ConstantExpression(new Integer(2)))),
                            parser.parse("2 + -2"));

        Assert.assertEquals("Parser fails for embedded unary expressions",
                            new BinaryOperatorExpression(constant(2),
                                                         "+",
                                                         constant(2)),
                            parser.parse("+2 + +2"));

        Assert.assertEquals("Parser fails for embedded unary expressions",
                            new BinaryOperatorExpression(constant(2),
                                                         "-",
                                                         constant(2)),
                            parser.parse("+2 - +2"));

        Assert.assertEquals("Parser fails for embedded unary expressions",
                            new UnaryMinusExpression(new BinaryOperatorExpression(constant(2),
                                                                                  "+",
                                                                                  constant(2))),
                            parser.parse("-(2+2)"));

        Assert.assertEquals("Parser fails unary minus with sub-expressions",
                            new UnaryMinusExpression(new ConstantExpression(new Integer(4))),
                            parser.parse("-(4)"));
    }

    public void testConstantEquality()
        throws Exception
    {
        Expression exp1 = new ConstantExpression(new Integer(1));
        Expression exp2 = new ConstantExpression(new Integer(1));

        Assert.assertEquals("", exp1, exp2);
    }

    public void testTrailingJunk()
        throws Exception
    {
        try
        {
            parser.parse("2+2)");

            Assert.fail("Parser allows trailing junk after expression");
        }
        catch (ParseException pe)
        {
            // Expected
        }
    }

    /**
     * Returns a Test which will execute all the test methods found
     * in this class.<p>
     *
     * @return A Test for JUnit to run.
     */
    public static Test suite()
    {
        TestSuite ts = new TestSuite(UnitTest.class);

        return ts;
    }

    /**
     * Creates a new UnitTest to execute the specified test.<p>
     *
     * @param name The name of the test method to run.
     */
    public UnitTest(String name)
    {
        super(name);
    }
}
